# Facial Recognition using the Person Sensor, Raspberry Pi Pico RP2040,
# and Pimoroni Pico Explorer Base
# Code for Instructable by CLClab306 01/26/2023
# This program was adapted from code written for the Person Sensor by Pete Warden.
# It also includes code snippets adapted from Adafruit's CircuitPython guides.
# For a complete technical description of Useful Sensor's Person Sensor:
# go to https://github.com/usefulsensors/person_sensor_docs

import asyncio
import board
import keypad
import digitalio
from digitalio import DigitalInOut, Direction, Pull
import bitmaptools
import busio
import terminalio
import displayio
import struct
from adafruit_display_text import label
from adafruit_st7789 import ST7789
import neopixel
import pwmio

# Edit the next line to provide different labels for ID#'s 0-7
names = ["A", "B", "C", "D", "E", "F", "G", "H", "-1"]
# names = ["Carter", "Reagan", "Bush1", "Clinton", "Bush2", "Obama", "Trump", "Biden", "-1"]
# names = ["Booker", "Hirono", "Ernst", "Harris", "Scott", "Murray", "Warnock", "Duckworth", "-1"]

criterion = 97      # ID% criterion for turning on the piezo and built-in LED

# Attach a jumper wire between GP0 & AUDIO on the Pico Explorer Base for builtin piezo
buzzer = pwmio.PWMOut(board.GP0, frequency=660, duty_cycle=0, variable_frequency=True)

# turn on optional light ring
pixel_pin = board.GP1
num_pixels = 16
ORDER = neopixel.GRB
pixels = neopixel.NeoPixel(pixel_pin, num_pixels, brightness=0.5, auto_write=False, pixel_order=ORDER)
pixels.fill((255, 255, 255))
pixels.show()

led = DigitalInOut(board.LED)                   # Pico built-in LED
led.direction = digitalio.Direction.OUTPUT

# Release any resources currently in use for the displays
displayio.release_displays()

# define pins for ST7789 LCD on Pico Explorer
tft_cs = board.GP17
tft_dc = board.GP16
spi_mosi = board.GP19
spi_clk = board.GP18
spi = busio.SPI(spi_clk, spi_mosi)
display_bus = displayio.FourWire(spi, command=tft_dc, chip_select=tft_cs)
display = ST7789(display_bus, width=240, height=240, rowstart=80, rotation=180)

# The person sensor has the I2C ID of hex 62, or decimal 98.
PERSON_SENSOR_I2C_ADDRESS = 0x62

# We will be reading raw bytes over I2C, and we'll need to decode them into
# data structures. These strings define the format used for the decoding, and
# are derived from the layouts defined in the Person Sensor Developer Guide.
PERSON_SENSOR_I2C_HEADER_FORMAT = "BBH"
PERSON_SENSOR_I2C_HEADER_BYTE_COUNT = struct.calcsize(PERSON_SENSOR_I2C_HEADER_FORMAT)

PERSON_SENSOR_FACE_FORMAT = "BBBBBBbB"
PERSON_SENSOR_FACE_BYTE_COUNT = struct.calcsize(PERSON_SENSOR_FACE_FORMAT)

PERSON_SENSOR_FACE_MAX = 4
PERSON_SENSOR_RESULT_FORMAT = (
    PERSON_SENSOR_I2C_HEADER_FORMAT
    + "B"
    + PERSON_SENSOR_FACE_FORMAT * PERSON_SENSOR_FACE_MAX
    + "H"
)
PERSON_SENSOR_RESULT_BYTE_COUNT = struct.calcsize(PERSON_SENSOR_RESULT_FORMAT)

# How long to pause between sensor polls.
PERSON_SENSOR_DELAY = 0.1

# The Pico doesn't support board.I2C(), so check before calling it. If it isn't
# present then we assume we're on a Pico and call an explicit function.
try:
    i2c = board.I2C()
except:
    i2c = busio.I2C(sda=board.GP20, scl=board.GP21)

# Wait until we can access the bus.
while not i2c.try_lock():
    pass

# Enable the continuous mode, the ID model & store recognized IDs
i2c.writeto(0x62, bytearray([0x01, 0x01]))
i2c.writeto(0x62, bytearray([0x02, 0x01]))
i2c.writeto(0x62, bytearray([0x05, 0x01]))

# Set up the Pico Explorer display
if display:
    bitmap = displayio.Bitmap(display.width, display.height, 2)
    palette = displayio.Palette(2)
    palette[0] = 0x000000
    palette[1] = 0xFFFF00  # yellow
    tile_grid = displayio.TileGrid(bitmap, pixel_shader=palette)
    group = displayio.Group()
    group.append(tile_grid)
    display.show(group)

# set label positions on the display
header_label = label.Label(font=terminalio.FONT, color=0x00FFFF, text="CLClab306", scale=2)
header_label.anchor_point = (0, 0)
header_label.anchored_position = (66, 0)
group.append(header_label)

ID_label = label.Label(font=terminalio.FONT, color=0x00FF00, text="ID", scale=5)
ID_label.anchor_point = (0, 0)
ID_label.anchored_position = (20, 40)
group.append(ID_label)

idConf_label = label.Label(font=terminalio.FONT, text="IDconf", scale=4)
idConf_label.anchor_point = (0, 0)
idConf_label.anchored_position = (20, 120)
group.append(idConf_label)

next_ID = 0         # Calibration ID #s range from 0-7; always start with #0
btn_label = label.Label(font=terminalio.FONT, text="Next ID# = {}".format(next_ID), scale=3)
btn_label.anchor_point = (0, 0)
btn_label.anchored_position = (20, 200)
group.append(btn_label)

# This is the interrupt_task used for getting button inputs from the Explorer
async def catch_buttons(clear, ID_dec, calibrate, ID_inc):
    global next_ID
    with keypad.Keys((clear, ID_dec, calibrate, ID_inc), value_when_pressed=False) as keys:
        while True:
            event = keys.events.get()
            if event and event.pressed:
                if event.key_number == 0:
                    print("\nClear")
                    i2c.writeto(0x62, bytearray([0x06]))
                    next_ID = 0
                if event.key_number == 1:
                    print("\nDecrement ID#")
                    next_ID -= 1
                    if next_ID < 0:
                        next_ID = 7
                if event.key_number == 2:
                    print("\nCALIBRATE ID# = {}".format(next_ID))
                    i2c.writeto(0x62, bytearray([0x04, next_ID]))
                    INC()
                if event.key_number == 3:
                    print("\nIncrement ID#")
                    INC()
                btn_label.text = "Next ID# = {}".format(next_ID)
                print("Next ID# = {}".format(next_ID))
            pixels.show()   # refresh the NeoPixels if needed
            await asyncio.sleep(0)

def INC():
    global next_ID
    next_ID += 1
    if next_ID > 7:
        next_ID = 0

# this is the sensor_task; it loops continously retrieving data from the Person Sensor
async def person_sensor():
    # Keep looping and reading the person sensor results.
    while True:
        # Get data from person_sensor
        read_data = bytearray(PERSON_SENSOR_RESULT_BYTE_COUNT)
        i2c.readfrom_into(PERSON_SENSOR_I2C_ADDRESS, read_data)
        offset = 0
        (pad1, pad2, payload_bytes) = struct.unpack_from(
            PERSON_SENSOR_I2C_HEADER_FORMAT, read_data, offset
        )
        offset = offset + PERSON_SENSOR_I2C_HEADER_BYTE_COUNT

        (num_faces) = struct.unpack_from("B", read_data, offset)
        num_faces = int(num_faces[0])
        offset = offset + 1

        faces = []
        for i in range(num_faces):
            (
                box_confidence,
                box_left,
                box_top,
                box_right,
                box_bottom,
                id_confidence,
                id,
                is_facing,
            ) = struct.unpack_from(PERSON_SENSOR_FACE_FORMAT, read_data, offset)
            offset = offset + PERSON_SENSOR_FACE_BYTE_COUNT
            face = {
                "box_confidence": box_confidence,
                "box_left": box_left,
                "box_top": box_top,
                "box_right": box_right,
                "box_bottom": box_bottom,
                "id_confidence": id_confidence,
                "id": id,
                "is_facing": is_facing,
            }
            faces.append(face)
        checksum = struct.unpack_from("H", read_data, offset)
        # Uncomment the next line to show more data in Serial Console
        # print(num_faces, faces)

        if num_faces > 0:
            main_face = faces[0]        # this is the largest (closest) face
            if main_face["is_facing"]:
                i = "Yes"
            else:
                i = "No"
            print("\nIs Facing = {}   Box% = {}".format(i, main_face["box_confidence"]), end="")

        # Facial recogntion is only applied to the largest face
        if num_faces > 0 and main_face["is_facing"]:
            id0 = main_face["id"]
            ID_label.text = str(names[id0])
            id0_conf = main_face["id_confidence"]
            idConf_label.text = str(id0_conf)
            print("   ID = {}   ID% = {}".format(names[id0], id0_conf), end="")
            if id0_conf >= criterion:
                buzzer.duty_cycle = 2 ** 15  # turn piezo ON when ID confidence is very high
                '''
                The following if/else code does something only when
                a specific ID# is detected with very high confidence; in this case
                Pico's built-in LED is turned on for ID# = 0 only
                '''
                if id0 == 0:
                    led.value = True
                else:
                    led.value = False
            else:
                led.value = False
                buzzer.duty_cycle = 0
        else:
            ID_label.text = ""
            idConf_label.text = ""
            led.value = False
            buzzer.duty_cycle = 0
        if num_faces == 0:
            ID_label.text = ""
            idConf_label.text = ""
            led.value = False
            buzzer.duty_cycle = 0

        # If the board has a display, draw rectangles for the face boxes, and a
        # diagonal cross if the person is facing the sensor.
        if display:
            bitmaptools.fill_region(bitmap, 0, 0, display.width, display.height, 0)
            for face in faces:
                box_left = face["box_left"]
                box_top = face["box_top"]
                box_right = face["box_right"]
                box_bottom = face["box_bottom"]
                x0 = int(((256-box_left) * display.width) / 256)
                y0 = int((box_top * display.height) / 256)
                x1 = int(((256-box_right) * display.width) / 256)
                y1 = int((box_bottom * display.height) / 256)
                # Uncomment the following lines to see bounding box coordinates in serial console
                # # print(x0, y0, x1, y1)
                # print(display.width, display.height)
                bitmaptools.draw_line(bitmap, x0, y0, x1, y0, 1)
                bitmaptools.draw_line(bitmap, x1, y0, x1, y1, 1)
                bitmaptools.draw_line(bitmap, x1, y1, x0, y1, 1)
                bitmaptools.draw_line(bitmap, x0, y1, x0, y0, 1)
                if face["is_facing"]:
                    bitmaptools.draw_line(bitmap, x0, y0, x1, y1, 1)
                    bitmaptools.draw_line(bitmap, x1, y0, x0, y1, 1)
        await asyncio.sleep(PERSON_SENSOR_DELAY)


async def main():
    interrupt_task = asyncio.create_task(catch_buttons(board.GP12, board.GP13, board.GP14, board.GP15))
    sensor_task = asyncio.create_task(person_sensor())
    await asyncio.gather(interrupt_task, sensor_task)

asyncio.run(main())
